==> coventry-0.8.1.tar.gz: coventry-0.8.1/utils/pbx-registry.cpp <==
/*
 * Copyright (C) 2022 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <moderncli/keyfile.hpp>
#include <moderncli/print.hpp>
#include <moderncli/strings.hpp>
#include <moderncli/filesystem.hpp>
#include <moderncli/args.hpp>

#include <unistd.h>
#include <fcntl.h>
#include <mqueue.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "server.hpp"
#include "ipc.hpp"

namespace {
args::option clearing('c',  "--clear",      "clear users registry keys");
args::option exists('e', "--exists",        "test extension exists");
args::option invite('i', "--invite",        "extension supports invite");
args::option helpflag('h',  "--help",       "display this list");
args::option message('m', "--message",      "extension supports message");
args::option quiet('q',     "--quiet",      "suppress all output");
args::option althelp('?',   nullptr, nullptr);

std::string keyname;
auto count = 0U;
auto extension = 0;

class ipc_error final : public std::exception {
public:
    explicit ipc_error(const char *text, int code) noexcept : msg(text), err(code) {}

    auto what() const noexcept -> const char * override {
        return msg;
    }

    auto exit() const noexcept {
        return err;
    }

private:
    const char *msg;
    int err;
};

auto get_value(std::string_view text) -> int {
    int value = 0;
    bool negative = false;

    if(!text.empty() && text.front() == '-') {
        negative = true;
        text.remove_prefix(1);
    }

    for(auto ch : text) {
        if(ch < '0' || ch > '9')
            return 0;
        if(value > 99999999)
            return 0;
        value *= 10;
        value += (ch - '0');
    }

    if(negative)
        return -value;

    return value;
}

auto set_string(char *cp, size_t max, std::string_view view) -> size_t {
    assert(cp != nullptr);

    auto size = view.size();
    auto dp = view.data();
    if(size >= max)
        size = max - 1;

    max = size;
    while(max--)
        *(cp++) = *(dp++);

    *cp = 0;
    return size;
}

auto set_lower(char *cp, size_t max, std::string_view view) -> size_t {
    assert(cp != nullptr);

    auto size = view.size();
    auto dp = view.data();
    if(size >= max)
        size = max - 1;

    max = size;
    while(max--)
        *(cp++) = static_cast<char>(tolower(*(dp++)));

    *cp = 0;
    return size;
}
} // end namespace

auto main(int argc, const char **argv) -> int { // NOLINT
    auto exit_code = 0;
    try {
        count = args::parse(argc, argv);
        auto list = args::list();
        if(is(helpflag) || is(althelp) || count < 1 || count > 3 || ((is(exists) || is(message) || is(invite)) && count != 1)) {
            args::help({"pbx-registry [options] extension [key [value]]"}, {
                "pbx-registry " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com",
                "Examine registry, see and modify user keys"
            });
            return 0;
        }

        fsys::current_path(SERVER_PREFIX_PATH);
        keyfile keys({SERVER_CONFIG_FILE, "dynamic.conf", "state.conf", "custom.conf", "registry.keys"});
        extension = get_value(list[0]);
        if(extension < 10 || extension > 89)
            throw std::runtime_error(format("{}: invalid extension", extension));
        // on throw undefined extension...
        if(!keys.exists(std::to_string(extension)))
            throw std::runtime_error(format("{}: undefined extension", extension));

        // test flags...
        if(is(invite) || is(message)) {
            auto rshm = shm_open(SERVER_IPCPATH ".registry", O_RDONLY, 0660);
            if(rshm < 0)
                throw ipc_error("failed", 4);
            struct stat rino{};
            if(fstat(rshm, &rino) < 0)
                throw ipc_error("failed", 4);
            auto rmax = (rino.st_size - sizeof(pbx_sys)) / sizeof(pbx_reg);
            auto rmem = mmap(nullptr, sizeof(pbx_reg) * rmax + sizeof(pbx_sys), PROT_READ, MAP_SHARED, rshm, 0);
            auto reg = static_cast<pbx_reg*>(rmem);
            auto sys = reinterpret_cast<const pbx_sys *>(&reg[rmax]);
            if(extension < sys->first || extension >= static_cast<int>(sys->first + rmax))
                throw ipc_error("offline", 1);

            const auto& user = reg[extension - sys->first];
            if(!user.flags.message && is(message))
                throw ipc_error("unsupported", 2);
            if(!user.flags.invite && is(invite))
                throw ipc_error("unsupported", 2);
            return 0;
        }

        if(is(exists))
            return 0;

        pbx_msg msg{.type = PBX_ACTIVE, .ver = PBX_VERSION};
        msg.body.registry.ext = extension;
        if(count > 1)
            keyname = lower_case(std::string{list[1]});
        if(is(clearing))
            msg.type = PBX_CLEARS;
        else if(count == 3) {
            set_lower(msg.body.registry.key, sizeof(msg.body.registry.key) - 1, list[1]);
            set_string(msg.body.registry.value, sizeof(msg.body.registry.value) - 1, list[2]);
            msg.type = PBX_ASSIGN;
        }
        auto mq = mq_open(SERVER_IPCPATH, O_WRONLY, 0660, nullptr);
        if(mq == MQD_INVALID)
            throw ipc_error("mqueue failed", 4);
        if(mq_send(mq, reinterpret_cast<const char *>(&msg), sizeof(msg), 0) < 0)
            throw ipc_error("send failed", 4);

        if(!is(quiet)) {
            if(is(clearing))
                print("user {} keys cleared\n", extension);
            else if(count == 3)
                print("user {} key set\n", extension);
            else if(count == 1)
                print("user {} valid\n", extension);
        }
        if(count == 2) {
            auto user = keys.at(std::to_string(extension));
            auto common = keys["common"];
            auto value = std::string();
            try {
                value = user.at(keyname);
            }
            catch(...) {
                try {
                    value = common.at(keyname);
                }
                catch(...) {}
            }
            print("{}\n", value);
        }
    }
    catch(const bad_arg& e) {
        die(-1, "Usage: pbx-registry [options] extensions [key [value]]\nerror: {}\n", e.what());
    }
    catch(const ipc_error& e) {
        exit_code = e.exit();
        if(!is(quiet))
            die(exit_code, "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: unknown error\n");
    }
    return exit_code;
}

==> coventry-0.8.6.tar.gz: coventry-0.8.6/utils/pbx-registry.cpp <==
/*
 * Copyright (C) 2022 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "server.hpp"       // IWYU pragma: keep
#include "ipc.hpp"

#include <moderncli/process.hpp>
#include <moderncli/keyfile.hpp>
#include <moderncli/strings.hpp>
#include <moderncli/filesystem.hpp>
#include <moderncli/args.hpp>

#include <mqueue.h>
#include <sys/stat.h>

namespace {
args::option clearing('c',  "--clear",      "clear users registry keys");
args::option exists('e', "--exists",        "test extension exists");
args::option invite('i', "--invite",        "extension supports invite");
args::option helpflag('h',  "--help",       "display this list");
args::option message('m', "--message",      "extension supports message");
args::option quiet('q',     "--quiet",      "suppress all output");
args::option althelp('?',   nullptr, nullptr);

std::string keyname;
auto count = 0U;
auto extension = 0;

class ipc_error final : public std::exception {
public:
    explicit ipc_error(const char *text, int code) noexcept : msg(text), err(code) {}

    auto what() const noexcept -> const char * override {
        return msg;
    }

    auto exit() const noexcept {
        return err;
    }

private:
    const char *msg;
    int err;
};

auto get_value(std::string_view text) -> int {
    int value = 0;
    bool negative = false;

    if(!text.empty() && text.front() == '-') {
        negative = true;
        text.remove_prefix(1);
    }

    for(auto ch : text) {
        if(ch < '0' || ch > '9')
            return 0;
        if(value > 99999999)
            return 0;
        value *= 10;
        value += (ch - '0');
    }

    if(negative)
        return -value;

    return value;
}

auto set_string(char *cp, size_t max, std::string_view view) -> size_t {
    assert(cp != nullptr);

    auto size = view.size();
    auto dp = view.data();
    if(size >= max)
        size = max - 1;

    max = size;
    while(max--)
        *(cp++) = *(dp++);

    *cp = 0;
    return size;
}

auto set_lower(char *cp, size_t max, std::string_view view) -> size_t {
    assert(cp != nullptr);

    auto size = view.size();
    auto dp = view.data();
    if(size >= max)
        size = max - 1;

    max = size;
    while(max--)
        *(cp++) = static_cast<char>(tolower(*(dp++)));

    *cp = 0;
    return size;
}
} // end namespace

auto main(int argc, const char **argv) -> int {
    using namespace process;

    auto exit_code = 0;
    try {
        count = args::parse(argc, argv);
        auto list = args::list();
        if(is(helpflag) || is(althelp) || count < 1 || count > 3 || ((is(exists) || is(message) || is(invite)) && count != 1)) {
            args::help({"pbx-registry [options] extension [key [value]]"}, {
                "pbx-registry " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com",
                "Examine registry, see and modify user keys"
            });
            return 0;
        }

        fsys::current_path(SERVER_PREFIX_PATH);
        keyfile keys({SERVER_CONFIG_FILE, "dynamic.conf", "state.conf", "custom.conf", "registry.keys"});
        extension = get_value(list[0]);
        if(extension < 10 || extension > 89)
            throw std::runtime_error(format("{}: invalid extension", extension));
        // on throw undefined extension...
        if(!keys.exists(std::to_string(extension)))
            throw std::runtime_error(format("{}: undefined extension", extension));

        // test flags...
        if(is(invite) || is(message)) {
            auto rshm = shm_open(SERVER_IPCPATH ".registry", O_RDONLY, 0660);
            if(rshm < 0)
                throw ipc_error("failed", 4);
            struct stat rino{};
            if(fstat(rshm, &rino) < 0)
                throw ipc_error("failed", 4);
            auto rmax = (rino.st_size - sizeof(pbx_sys)) / sizeof(pbx_reg);
            auto rmem = map_t(rshm, rino.st_size, false);
            if(!is(rmem))
                throw ipc_error("failed", 4);
            auto reg = static_cast<pbx_reg*>(rmem.data());
            auto sys = reinterpret_cast<const pbx_sys *>(&reg[rmax]);
            if(extension < sys->first || extension >= static_cast<int>(sys->first + rmax))
                throw ipc_error("offline", 1);

            const auto& user = reg[extension - sys->first];
            if(!user.flags.message && is(message))
                throw ipc_error("unsupported", 2);
            if(!user.flags.invite && is(invite))
                throw ipc_error("unsupported", 2);
            return 0;
        }

        if(is(exists))
            return 0;

        pbx_msg msg{.type = PBX_ACTIVE, .ver = PBX_VERSION};
        msg.body.registry.ext = extension;
        if(count > 1)
            keyname = lower_case(std::string{list[1]});
        if(is(clearing))
            msg.type = PBX_CLEARS;
        else if(count == 3) {
            set_lower(msg.body.registry.key, sizeof(msg.body.registry.key) - 1, list[1]);
            set_string(msg.body.registry.value, sizeof(msg.body.registry.value) - 1, list[2]);
            msg.type = PBX_ASSIGN;
        }
        auto mq = mq_open(SERVER_IPCPATH, O_WRONLY, 0660, nullptr);
        if(mq == MQD_INVALID)
            throw ipc_error("mqueue failed", 4);
        if(mq_send(mq, reinterpret_cast<const char *>(&msg), sizeof(msg), 0) < 0)
            throw ipc_error("send failed", 4);

        if(!is(quiet)) {
            if(is(clearing))
                print("user {} keys cleared\n", extension);
            else if(count == 3)
                print("user {} key set\n", extension);
            else if(count == 1)
                print("user {} valid\n", extension);
        }
        if(count == 2) {
            auto user = keys.at(std::to_string(extension));
            auto common = keys["common"];
            auto value = std::string();
            try {
                value = user.at(keyname);
            }
            catch(...) {
                try {
                    value = common.at(keyname);
                }
                catch(...) {}   // NOLINT
            }
            print("{}\n", value);
        }
    }
    catch(const bad_arg& e) {
        die(-1, "Usage: pbx-registry [options] extensions [key [value]]\nerror: {}\n", e.what());
    }
    catch(const ipc_error& e) {
        exit_code = e.exit();
        if(!is(quiet))
            die(exit_code, "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: unknown error\n");
    }
    return exit_code;
}
